


#ifndef UTILS_H
#define UTILS_H


#define DEBUG 

#ifdef __APPLE__
#define ok_qsort_r qsort_r
#else
typedef int		 cmp_t(void *, const void *, const void *);
extern void ok_qsort_r(void *a, size_t n, size_t es, void *thunk, cmp_t *cmp);
#endif

#define MIN(a,b) (a < b ? a : b)
#define MAX(a,b) (a > b ? a : b)

#define RANGE(x, a, b) (MIN(MAX(x, a), b))

#define MGET(m, i, j) (m->data[(i)*m->tda+(j)])
#define MSET(m, i, j, x) m->data[(i)*m->tda+(j)] = x 
#define MIGET gsl_matrix_int_get
#define MISET gsl_matrix_int_set

#define MINC(m, i, j, c) MSET(m, i, j, MGET(m, i, j) + c)
#define MROWS(m) (m->size1)
#define MCOLS(m) (m->size2)

#define SQR(x) ((x)*(x))

#define VGET(v, i) (v->data[(i)])
#define VSET(v, i, x) v->data[(i)] = (x)
#define VIGET gsl_vector_int_get
#define VISET gsl_vector_int_set
#define ROW gsl_matrix_row
#define COL gsl_matrix_col

#define VLEN(X) (sqrt(X[0] * X[0] + X[1] * X[1] + X[2] * X[2]))
#define DELTA(I, J) (I == J ? 1 : 0)

#define MSUN_TO_INT(M) (M * K2)
#define INT_TO_MSUN(M) (M / K2)

#define MJUP_TO_INT(M) (M * MJUP / MSUN * K2)
#define INT_TO_MJUP(M) (M / K2 * MSUN / MJUP)

#define MSUN_TO_MJUP(M) (M * MJUP / MSUN)
#define MJUP_TO_MSUN(M) (M * MSUN / MJUP)

#define GM_TO_INT(M) (M / MSUN * K2)
#define INT_TO_GM(M) (M / K2 * MSUN)

#define DAY_TO_YEAR(T) (T * DAY / YEAR)
#define YEAR_TO_DAY(T) (T / DAY * YEAR)

#define CM_TO_INT(L) (L / AU)
#define INT_TO_CM(L) (L * AU)

#define AUPDAY_TO_MPS(V) (V * AU / 100 / DAY)
#define MPS_TO_AUPDAY(V) (V * DAY / (AU / 100))

#define sqr(x) ((x)*(x))
#define TO_DEG(ANGLE) (DEGRANGE(ANGLE / TWOPI * 360.))
#define TO_RAD(ANGLE) (RADRANGE(ANGLE / 360. * TWOPI))

#define MATRIX_MEMCPY(dest, src) memcpy(dest->data, src->data, sizeof(double)*src->size1*src->size2)

#define K_GETSET_H(name, Name, type) \
        void K_set##Name(ok_kernel* k, type name);\
        type K_get##Name(ok_kernel* k);

#define K_GETSET_C(name, Name, type) \
        void K_set##Name(ok_kernel* k, type name) { k-> name = name; K_validate(k); k->flags |= NEEDS_SETUP; } \
        type K_get##Name(ok_kernel* k) { return k-> name; }

#define K_GETSET_FREE_C(name, Name, type, freefcn) \
        void K_set##Name(ok_kernel* k, type name) { if (k-> name == name) return; freefcn(k->name); k-> name = name;  } \
        type K_get##Name(ok_kernel* k) { return k-> name; }


#define K_GETSETW_C(name, Name, type, where) \
        void K_set##Name(ok_kernel* k, type name) { where = name; } \
        type K_get##Name(ok_kernel* k) { return where; }

#define K_GET_H(name, Name, type) \
        type K_get##Name(ok_kernel* k);

#define K_GET_C(name, Name, type) \
        type K_get##Name(ok_kernel* k) { return k-> name; }


typedef struct {
    void* data;
    int tag[5];
} ok_tagdata;


double DEGRANGE(double angle);
double RADRANGE(double angle);


void ok_sort_small_matrix(gsl_matrix* matrix, const int column);
void ok_sort_matrix(gsl_matrix* matrix, const int column);
void ok_rsort_matrix(gsl_matrix* matrix, const int column);

void ok_fprintf_matrix(gsl_matrix* matrix, FILE* file, const char* format);
void ok_fprintf_vector(gsl_vector* v, FILE* file, const char* format);
void ok_fprintf_matrix_int(gsl_matrix_int* matrix, FILE* file, const char* format);
void ok_fprintf_vector_int(gsl_vector_int* v, FILE* file, const char* format);
void ok_fprintf_buf(double** buf, FILE* file, const char* format, int rows, int columns);

void ok_fscanf_matrix(gsl_matrix* matrix, FILE* file);
void ok_fscanf_matrix_int(gsl_matrix_int* matrix, FILE* file);
void ok_fscanf_vector(gsl_vector* matrix, FILE* file);
void ok_fscanf_vector_int(gsl_vector_int* matrix, FILE* file);


bool ok_save_matrix(gsl_matrix* matrix, FILE* fid, const char* format);
bool ok_save_matrix_bin(gsl_matrix* matrix, FILE* fid);
bool ok_save_buf(double** matrix, FILE* fid, const char* format, int rows, int cols);
bool ok_save_buf_bin(double** matrix, FILE* fid, int rows, int cols);

gsl_matrix* ok_read_table(FILE* fid);

gsl_vector* ok_vector_resize(gsl_vector* v, int len);
gsl_matrix* ok_matrix_resize(gsl_matrix* v, int rows, int cols);
gsl_vector* ok_vector_resize_pad(gsl_vector* v, int len, double pad);
gsl_matrix* ok_matrix_resize_pad(gsl_matrix* v, int rows, int cols, double pad);

gsl_vector_int* ok_vector_int_resize(gsl_vector_int* v, int len);
gsl_matrix_int* ok_matrix_int_resize(gsl_matrix_int* v, int rows, int cols);

gsl_matrix* ok_matrix_copy(const gsl_matrix* src);
gsl_matrix_int* ok_matrix_int_copy(const gsl_matrix_int* src);
gsl_vector* ok_vector_copy(const gsl_vector* src);
gsl_vector_int* ok_vector_int_copy(const gsl_vector_int* src);
gsl_matrix* ok_buf_to_matrix(double** buf, int rows, int cols);

gsl_matrix* ok_matrix_remove_row(gsl_matrix* m, int row);
gsl_matrix* ok_matrix_remove_column(gsl_matrix* m, int col);
gsl_matrix_int* ok_matrix_int_remove_row(gsl_matrix_int* m, int row);
gsl_matrix_int* ok_matrix_int_remove_column(gsl_matrix_int* m, int col);
gsl_vector* ok_vector_remove(gsl_vector* m, int idx);
gsl_vector_int* ok_vector_int_remove(gsl_vector_int* m, int idx);

void ok_bootstrap_matrix(const gsl_matrix* matrix, gsl_matrix* out, const int sortcol, gsl_rng* r);
void ok_bootstrap_matrix_mean(const gsl_matrix* matrix, int timecol, int valcol, gsl_matrix* out, gsl_rng* r);

gsl_matrix* ok_matrix_filter(gsl_matrix* matrix, const int column, const double filter);
gsl_matrix* ok_matrix_buf_filter(double** matrix, const int rows, const int columns, const int column, const double filter);

int ok_bsearch(double* v, double val, int len);

double ok_average_angle(const double* v, const int length, const bool isRadians);
double ok_median_angle(const double* v, const int length, const bool isRadians);
double ok_stddev_angle(const double* v, const int length, const bool isRadians);
char* ok_str_copy(const char* src);
char* ok_str_cat(const char* a1, const char* a2);

void ok_avevar(const double* v, int len, double* ave, double* var);

#endif



#ifndef types_h
#define types_h

#define INVALID_NUMBER (NAN)
#define IS_INVALID(x) (isnan(x))

#define VERSION 1.1017

#define T_RV 0
#define T_PHOTO 1
#define T_TIMING 2

#define DATA_SIZE 11
#define T_TIME 0
#define T_VAL 1
#define T_ERR 2
#define T_TDS_PLANET 3
#define T_FLAG 6
#define T_SVAL 7
#define T_PRED 8
#define T_SET 9
#define T_GUI 10

#define X 1
#define Y 2
#define Z 3
#define VX 4
#define VY 5
#define VZ 6

#define STAT_MEAN 0
#define STAT_MEDIAN -1
#define STAT_STDDEV -2
#define STAT_QUART_25 25
#define STAT_QUART_50 50
#define STAT_QUART_75 75

#define ENSURE_SIZE(matrix, rows, columns) assert((matrix->size1 == rows) && (matrix->size2 == columns))

#define KEPLER 0
#define RK5 1
#define RK8 2
#define SWIFT_BS 3
#define SWIFT_WHM 4
#define SWIFT_RMVS3 5

#define AU  1.4959787e13
#define MSUN 1.98892e33
#define MJUP 1.8986e30
#define RJUP 7.1e9
#define GGRAV 6.67384e-8

#define DAY 8.64e4
#define TWOPI (2.*M_PI)
#define YEAR 31556926.

#define K2  ((GGRAV * MSUN * DAY * DAY) / (AU*AU*AU))

#define PER 0
#define MASS 1
#define MA 2
#define ECC 3
#define LOP 4
#define INC 5
#define NODE 6
#define RADIUS 7
#define ORD 8
#define SMA 13
#define SEMIAMP 14
#define ECCANOMALY 15
#define TRUEANOMALY 16

#define ELEMENTS_SIZE 13
#define ALL_ELEMENTS_SIZE 17

#define TREND 8
#define NOISE 9
#define PARAMS_SIZE 15

extern char * ok_orb_labels[ELEMENTS_SIZE];

#define OPT_EPS 0
#define OPT_ECC_LAST 1

#define PROGRESS_CONTINUE 0
#define PROGRESS_STOP 1

#define DONE -1

typedef struct ok_system {
        int nplanets;
            double epoch;
                    gsl_matrix* elements;

            double time;
            gsl_matrix* xyz;
            gsl_matrix* orbits;
        int flag;
} ok_system;

typedef struct ok_integrator_options {
    double abs_acc;
    double rel_acc;
    bool calc_elements;
                int (* force) (double, const double[], double[], void*);
    int (* jac) (double, const double[], double*, double[], void*);
            double** buffer;
    
} ok_integrator_options;

typedef struct ok_kernel ok_kernel;
typedef double(*ok_callback)(ok_kernel*);
typedef ok_system**(*ok_integrator)(ok_system*, const gsl_vector*, const ok_integrator_options*);
typedef int(*ok_minimizer)(ok_kernel*, int, double[]);
typedef int(*ok_progress)(int current, int max, void* state, const char* function);

struct ok_kernel {
        ok_system* system;
        double chi2;
        double chi2_rvs;
        double chi2_tts;
        double rms;
        double rms_tts;
        double jitter;
        int nsets;
        gsl_matrix** datasets;

            double** compiled;
        gsl_vector* params;
        gsl_vector* times;
        ok_system** integration;
        int integrationSamples;

        gsl_matrix_int* plFlags;
    gsl_vector_int* parFlags;
    gsl_matrix* plSteps;
    gsl_vector* parSteps;
    gsl_matrix** plRanges;
    gsl_vector** parRanges;

        double Mstar;

        int ndata;
    int nrvs;
    int ntts;
    int nrpars;
    
        int intMethod;
    ok_integrator_options* intOptions;

        ok_callback minfunc;

        unsigned int flags;

        gsl_rng* rng;

        void* tag;
    
        ok_progress progress;
};

typedef struct ok_list ok_list;

struct ok_list {
    ok_kernel** kernels;
    int size;
};

#endif



#define ACTIVE (1 << 1)
#define MINIMIZE (1 << 2)

#define NEEDS_COMPILE (1 << 1)
#define NEEDS_SETUP (1 << 2)
#define BOOTSTRAP_DATA (1 << 3)
#define RETAIN (1 << 4)
#define SHARE_FLAGS (1 << 5)
#define SHARE_STEPS (1 << 6)
#define SHARE_DATA (1 << 7)
#define FREED (1 << 8)
#define DUPLICATE (1 << 9)

#define SIMPLEX 0
#define DFMIN 1

extern double ok_default_steps[];

ok_kernel* K_alloc();
void K_free(ok_kernel* k);
ok_kernel* K_clone(ok_kernel* k);
ok_kernel* K_cloneFlags(ok_kernel* k, unsigned int shareFlags);

gsl_matrix* K_addDataFile(ok_kernel* k, const char* path, int type);
gsl_matrix* K_addDataTable(ok_kernel* k, gsl_matrix* rvtable, const char* name, int type);
void K_removeData(ok_kernel* k, int idx);
gsl_matrix* K_getData(ok_kernel* k, int idx);

double** K_compileData(ok_kernel* k);
gsl_matrix* K_getResiduals(ok_kernel* k);
gsl_matrix* K_combineData(ok_kernel* k, const int data_type);

void K_addPlanet(ok_kernel* k, const double elements[]);
void K_removePlanet(ok_kernel* k, int idx);
void K_setElement(ok_kernel* k, int row, int col, double value);
double K_getElement(ok_kernel* k, int row, int col);
void K_setElements(ok_kernel* k, gsl_matrix* elements);
gsl_matrix* K_getElements(ok_kernel* k);
void K_setElementFlag(ok_kernel* k, int row, int col, int value);
int K_getElementFlag(ok_kernel* k, int row, int col);
void K_setElementStep(ok_kernel* k, int row, int col, double value);
double K_getElementStep(ok_kernel* k, int row, int col);
void K_setElementRange(ok_kernel* k, int row, int col, double min, double max);
void K_getElementRange(ok_kernel* k, int row, int col, double* min, double* max);
gsl_matrix* K_getAllElements(ok_kernel* k);

K_GETSET_H(Mstar, Mstar, double)
        
K_GETSET_H(epoch, Epoch, double)
K_GET_H(chi2, Chi2, double)
K_GET_H(chi2_nr, Chi2_nr, double)
K_GET_H(rms, Rms, double)
K_GET_H(jitter, Jitter, double)
K_GET_H(compiled, Compiled, double**)
K_GETSET_H(flags, Flags, unsigned int)
K_GETSET_H(minfunc, MinFunc, ok_callback)
K_GETSET_H(plSteps, ElementSteps, gsl_matrix*)
K_GETSET_H(parSteps, ParSteps, gsl_vector*)
K_GETSET_H(plFlags, ElementFlags, gsl_matrix_int*)
K_GETSET_H(parFlags, ParFlags, gsl_vector_int*)

K_GETSET_H(intMethod, IntMethod, int)
K_GETSET_H(intOptions, IntOptions, ok_integrator_options*)

unsigned int K_getNplanets(ok_kernel* k);
unsigned int K_getNRV(ok_kernel* k);
unsigned int K_getNdata(ok_kernel* k);
void K_getRange(ok_kernel* k, double* from, double* to);



void K_setPars(ok_kernel* k, gsl_vector* pars);
gsl_vector* K_getPars(ok_kernel* k);
void K_setPar(ok_kernel* k, int idx, double val);
double K_getPar(ok_kernel* k, int idx);
void K_setParFlag(ok_kernel* k, int idx, int value);
int K_getParFlag(ok_kernel* k, int idx);
void K_setParStep(ok_kernel* k, int idx, double value);
double K_getParStep(ok_kernel* k, int idx);
void K_setParRange(ok_kernel* k, int idx, double min, double max);
void K_getParRange(ok_kernel* k, int idx, double* min, double* max);

bool K_save(ok_kernel* k, FILE* fid);
ok_kernel* K_load(FILE* fid, int skip);

void K_calculate(ok_kernel* k);
int K_minimize(ok_kernel* k, int algo, int maxiter, double params[]);
int K_1dminimize(ok_kernel* k, int algo, int maxiter, int row, int column, double params[]);

ok_system** K_integrate(ok_kernel* k, gsl_vector* times, ok_system** bag);
ok_system** K_integrateRange(ok_kernel* k, double from, double to, unsigned int samples, ok_system** bag);
void K_print(ok_kernel* k, FILE* f);
void K_save_old(ok_kernel* k, const char* stem);
void K_setSeed(ok_kernel* k, unsigned long int seed);

#define OK_SUCCESS 0
#define OK_NOCONV 1

extern double ok_min_distance;
extern ok_integrator_options defoptions;

ok_system* ok_alloc_system(int nplanets);

void ok_free_system(ok_system* system);

ok_system* ok_copy_system(const ok_system* orig);

void ok_resize_system(ok_system* system, int npnew);

void ok_el2cart(ok_system* system, gsl_matrix* xyz);

void ok_cart2el(ok_system* system, gsl_matrix* els, bool internal);

void ok_sort_elements(ok_system* system, int column);

void ok_setup(ok_system* system);

int ok_force(double t, const double y[], double f[], void *params);

int ok_jac(double t, const double y[], double *dfdy, 
    double dfdt[], void *params);

ok_system** ok_integrate(ok_system* initial, const gsl_vector* times, const ok_integrator_options* options, const int integrator,
        ok_system** bag);

double ok_get_rv(ok_system* sys);
gsl_matrix* ok_get_rvs(ok_system** sys, int len);

gsl_matrix* ok_get_xyzs(ok_system** bag, int len);

gsl_matrix* ok_get_els(ok_system** bag, int len, bool internal);

void ok_to_cm(ok_system* system, gsl_matrix* xyz);

void ok_to_star(ok_system* system, gsl_matrix* xyz);

int ok_find_closest_transit(ok_system* sys, const int pidx, const ok_integrator_options* options, const int intMethod, double eps, double* timeout);



#ifndef KL_H
#define	KL_H

#ifdef	__cplusplus
extern "C" {
#endif
    
    ok_list* KL_alloc(const int size);
    void KL_free(ok_list* list);

    gsl_vector* KL_getElements(const ok_list* kl, const int pl, const int el);
    gsl_vector* KL_getPars(const ok_list* kl, const int vo);
    gsl_matrix* KL_getElementsStats(const ok_list* kl, const int what);
    unsigned int KL_getSize(const ok_list* kl);
    void KL_removeAtIndex(ok_list* kl, const unsigned int idx);
    void KL_fprintf(const ok_list* kl, FILE* out, const char* fmt, const char* lfmt);
#ifdef	__cplusplus
}
#endif

#endif	



#ifndef MCMC_H
#define	MCMC_H

#ifdef	__cplusplus
extern "C" {
#endif

    
ok_list* K_mcmc(ok_kernel* k, int steps, int discard, double param[], gsl_matrix** diags);

#ifdef	__cplusplus
}
#endif

#endif	



#ifndef BOOTSTRAP_H
#define	BOOTSTRAP_H

#ifdef	__cplusplus
extern "C" {
#endif


    ok_list* K_bootstrap(ok_kernel* k, int trials, int warmup, int malgo, int miter, double mparams[]);
    
#ifdef	__cplusplus
}
#endif

#endif	



#ifndef PERIODOGRAM_H
#define	PERIODOGRAM_H

#ifdef	__cplusplus
extern "C" {
#endif

    
#define PS_TIME 0
#define PS_Z 1
#define PS_FAP 2
#define PS_Z_LS 3
#define PS_TAU 4
#define PS_SIZE 6
    
    typedef struct {
        double W;
        double zmax;
        double z2max;
        double fapmax;
        double z_fap_1;
        double z_fap_2;
        double z_fap_3;
        bool calc_z_fap;
        gsl_matrix* per;
        gsl_matrix* buf;
        gsl_vector* zm;
    } ok_periodogram_workspace;
    
    gsl_matrix* ok_periodogram_ls(const gsl_matrix* data, const unsigned int samples, const double Pmin, const double Pmax, const int method,
        unsigned int timecol, unsigned int valcol, unsigned int sigcol, ok_periodogram_workspace* p);

    gsl_matrix* ok_periodogram_boot(const gsl_matrix* data, const unsigned int trials, const unsigned int samples, 
        const double Pmin, const double Pmax, const int method,
        const unsigned int timecol, const unsigned int valcol, const unsigned int sigcol,
        gsl_rng* rng, ok_periodogram_workspace* p, ok_progress prog);

#ifdef	__cplusplus
}
#endif

#endif	

